import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import tetrisbean.*;

/** Class implements the game of tetris in a panel that can be dropped
 * in any swing frame or window (possibly applet also).<br>
 * Uses the excellent tetrisbean package created by Scott Clee,
 * IBM DeveloperWorks.<br>
 * NOTE: This panel should be added to the frame it sits in as a
 * keylistener - frame.addKeyListener(this panel reference).<br><br>
 * <b>Changes:</b>
 * <ul>
 * <li> 2002-05-27 - Initial release.
 * <li> 2002-12-07 - Added in a focus listener (for java 1.4+).
 * <li> 2002-12-09 - Added in keylistener to pause button (if under java 1.4+).
 *                   Also added in VK_ESC to work as pause (buttonless).
 * </ul>
 * @author Michael G. Slack
 * @author slack@attglobal.net
 * @version Version 1.02
*/
public class JTetrisPanel extends JPanel implements KeyListener
{
    private final static String S_START  = " :-) ";
    private final static String S_SCORE  = "Score: ";
    private final static String S_PAUSED = "< Paused >";
    private final static String S_OVER   = "Game Over";

    private TetrisGame     fGame     = new TetrisGame();
    private TetrisBoardGUI fBoardGUI = new TetrisBoardGUI();
    private NextPiece      nxtPc     = new NextPiece();
    private JLabel         lblScore  = new JLabel(S_SCORE+"0");
    private JPanel         pnlEast   = new JPanel();
    private JPanel         pnlBtns   = new JPanel();
    private JButton        btnStart  = new JButton("Start");
    private JButton        btnPause  = new JButton("Pause");
    private JButton        btnStop   = new JButton("Stop");
    private JLabel         lblMsg    = new JLabel(S_START, SwingConstants.CENTER);
    private boolean        paused    = false;

    // --------------------------- Constructors ---------------------------

    public JTetrisPanel()
    {
        super();
        // set it up
        initPanel();
    }

    public JTetrisPanel(boolean isDoubleBuffered)
    {
        super(isDoubleBuffered);
        // set it up
        initPanel();
    }

    /** Constructor to pass in a new layout manager.  LayoutManager is
     * ignored - uses a border layout for this panel only.
    */
    public JTetrisPanel(LayoutManager layout)
    {
        this();
    }

    /** Constructor to pass in a new layout manager along with double buffer
     * flag.  LayoutManager is ignored - uses a border layout only.
    */
    public JTetrisPanel(LayoutManager layout, boolean isDoubleBuffered)
    {
        this(isDoubleBuffered);
    }

    // -------------------------- Private Methods -------------------------

    /** Method used to initialize the panel with the buttons, game objects
     * and labels used.
    */
    private void initPanel()
    {
        setLayout(new BorderLayout()); // only layout for this panel
        // add components to game panel now
        add(lblScore, BorderLayout.NORTH);
        add(fBoardGUI, BorderLayout.CENTER);

        // create eastern panel (with layout and button panels)
        pnlEast.setLayout(new BorderLayout());
        pnlEast.add(nxtPc, BorderLayout.NORTH);
        pnlBtns.setLayout(new GridLayout(4, 1));
        pnlBtns.add(btnStart);
        btnPause.setEnabled(false);
        pnlBtns.add(btnPause);
        btnStop.setEnabled(false);
        pnlBtns.add(btnStop);
        pnlBtns.add(lblMsg);
        pnlEast.add(pnlBtns, BorderLayout.SOUTH);
        add(pnlEast, BorderLayout.EAST);

        // add button listeners
        btnStart.addActionListener(new ActionListener() {
                                       public void actionPerformed(ActionEvent evt) {
                                           btnPause.requestFocus(); // java 1.4+ fix
                                           fGame.startGame();
                                       }
                                   });
        btnPause.addActionListener(new ActionListener() {
                                       public void actionPerformed(ActionEvent evt) {
                                           paused = !paused;
                                           fGame.setPaused(paused);
                                           if (paused)
                                               lblMsg.setText(S_PAUSED);
                                           else
                                               lblMsg.setText(S_START);
                                       }
                                   });
        btnStop.addActionListener(new ActionListener() {
                                      public void actionPerformed(ActionEvent evt) {
                                          fGame.stopGame();
                                      }
                                  });

        // add listeners
        fGame.addBoardListener(fBoardGUI);
        fGame.addBoardListener(nxtPc);
        fGame.addScoreListener(new ScoreListener() {
                                   public void scoreChange(ScoreEvent e) {
                                       lblScore.setText(S_SCORE+e.getScore());
                                   }
                               });
        fGame.addGameListener(new GameListener() {
                                  public void gameChange(GameEvent e) {
                                      if (e.getType() == GameEvent.END) {
                                          lblMsg.setText(S_OVER);
                                          btnStart.setEnabled(true);
                                          btnPause.setEnabled(false);
                                          btnStop.setEnabled(false);
                                      }
                                      else if (e.getType() == GameEvent.START) {
                                          btnStart.setEnabled(false);
                                          btnPause.setEnabled(true);
                                          btnStop.setEnabled(true);
                                          lblMsg.setText(S_START);
                                      }
                                  }
                              });

        // add key listeners to pause button (java 1.4+ fix...)
        String ver = System.getProperty("java.version");
        if (ver.compareTo("1.4") > 0) { // 1.4.x version and above only
            btnPause.addKeyListener(this);
        }
    }

    /** Private method used to get the color of the piece from the shape of
     * the piece.  Moved out of TetrisBoardGUI.
    */
    private Color getPieceColor(int color)
    {
        Color result = null;

        switch (color) {
            case TetrisPiece.L_PIECE : result = new Color(24, 105, 198);  // turquoise
                                       break;
            case TetrisPiece.J_PIECE : result = new Color(206, 56, 173);  // purple
                                       break;
            case TetrisPiece.I_PIECE : result = new Color(41, 40, 206);   // blue
                                       break;
            case TetrisPiece.Z_PIECE : result = new Color(212, 0, 0);     // red
                                       break;
            case TetrisPiece.S_PIECE : result = new Color(82, 154, 16);   // green
                                       break;
            case TetrisPiece.O_PIECE : result = new Color(123, 121, 123); // gray
                                       break;
            case TetrisPiece.T_PIECE : result = new Color(156, 142, 8);   // yellow
                                       break;
        }

        return result;
    }

    /** Private method used to draw a piece of a tetris piece in the location
     * specified.  Moved from TetrisBoardGUI.
    */
    private void drawBlock(Graphics g, int x, int y, int width, int height)
    {
        g.fillRect(x, y, width, height);
        g.setColor(Color.black);
        g.drawRect(x, y, width, height);
    }

    // --------------------------- Public Methods -------------------------

    public void keyPressed(KeyEvent e)
    {
        switch (e.getKeyCode()) {
            case KeyEvent.VK_NUMPAD4 :
            case KeyEvent.VK_LEFT    : fGame.move(TetrisPiece.LEFT);
                                       break;
            case KeyEvent.VK_NUMPAD6 :
            case KeyEvent.VK_RIGHT   : fGame.move(TetrisPiece.RIGHT);
                                       break;
            case KeyEvent.VK_NUMPAD8 :
            case KeyEvent.VK_UP      : fGame.move(TetrisPiece.ROTATE);
                                       break;
            case KeyEvent.VK_NUMPAD2 :
            case KeyEvent.VK_DOWN    : fGame.move(TetrisPiece.DOWN);
                                       break;
            case KeyEvent.VK_SHIFT   : fGame.move(TetrisPiece.FALL);
                                       break;
            case KeyEvent.VK_ESCAPE  : btnPause.doClick();
        }
    }

    public void keyReleased(KeyEvent e)
    {
        // do nothing
    }

    public void keyTyped(KeyEvent e)
    {
        // do nothing
    }

    // ------------------------ Other Inner Classes -----------------------

    /**
     * This class displays the next piece that will be played in the game.
    */
    public class NextPiece extends JPanel implements BoardListener
    {
        public void boardChange(BoardEvent e)
        {
            if (e.getType() == BoardEvent.PIECE_ADDED) {
                repaint();
            }
        }

        public Dimension getPreferredSize()
        {
            return new Dimension(84, 84);
        }

        public void paintComponent(Graphics graphics)
        {
            int width  = getBounds().width;
            int height = getBounds().height;	

            // set up double buffering
            Image image = createImage(width, height);
            Graphics g = image.getGraphics();           

            // draw the next board piece
            TetrisPiece piece = fGame.getNextPiece();
            if (piece != null) {
                Point centre = piece.getCentrePoint();
                Point[] blocks = piece.getRelativePoints();
                Color color = getPieceColor(piece.getType());
                for (int count = 0; count < 4; count++) {
                    int x = centre.x + blocks[count].x + 1;
                    int y = centre.y + blocks[count].y + 1;

                    g.setColor(color);
                    drawBlock(g, (x * width / 4) + 1,
                                 (y * height / 4) + 1,
                                 (width / 4) - 1,
                                 (height / 4) - 1);
                }
            }

            g.setColor(Color.blue);
            g.drawRect(0, 0, width - 1, height - 1);
            graphics.drawImage(image, 0, 0, width, height, null);
        }
    }

    /**
     * This class gives a graphical representation of a TetrisBoard GUI.
     * - Copied from 'Scottris' version of Tetris game written by
     *   Scott Clee.
    */
    public class TetrisBoardGUI extends JPanel implements BoardListener
    {
        private TetrisBoard fBoard;
    
        public void boardChange(BoardEvent e)
        {
    	    fBoard = (TetrisBoard)e.getSource();   
            repaint();
        }

        public void paintComponent(Graphics graphics)
        {
            final int width  = getBounds().width;
            final int height = getBounds().height;	

            // Set up double buffering.
            final Image image = createImage(width, height);
            final Graphics g = image.getGraphics();           

            // Draw the board pieces if board not null.
            if (fBoard != null) {
                final int numCols = fBoard.getColumns();
                final int numRows = fBoard.getRows();

                for (int cols = 0; cols < numCols; cols++) {
                    for (int rows = 0; rows < numRows; rows++) {
                        final int piece = fBoard.getPieceAt(cols, rows);

                        if (piece != TetrisBoard.EMPTY_BLOCK) {
                            g.setColor(getPieceColor(piece));
                            drawBlock(g, (cols * width / numCols) + 1,
                                         (rows * height / numRows) + 1,
                                         (width / numCols) - 1,
                                         (height / numRows) - 1);
                        }
                    }
                }
            }	

            g.setColor(Color.blue);
            g.drawRect(0, 0, width - 1, height - 1);
            graphics.drawImage(image, 0, 0, width, height, null);
        }

        public Dimension getPreferredSize()
        {
            return new Dimension(200, 600);
        }
    }
}

